/*
 *  jDTAUS - DTAUS fileformat.
 *  Copyright (c) 2005 Christian Schulte <cs@schulte.it>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */
package org.jdtaus.core.monitor.ri;

import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Locale;
import java.util.Map;
import javax.swing.event.EventListenerList;
import org.jdtaus.core.container.ContainerFactory;
import org.jdtaus.core.container.ContainerInitializer;
import org.jdtaus.core.container.ContextFactory;
import org.jdtaus.core.container.ContextInitializer;
import org.jdtaus.core.container.Dependency;
import org.jdtaus.core.container.Implementation;
import org.jdtaus.core.container.ModelFactory;
import org.jdtaus.core.container.Properties;
import org.jdtaus.core.container.Property;
import org.jdtaus.core.container.PropertyException;
import org.jdtaus.core.logging.spi.Logger;
import org.jdtaus.core.monitor.Task;
import org.jdtaus.core.monitor.TaskEvent;
import org.jdtaus.core.monitor.TaskListener;
import org.jdtaus.core.monitor.spi.TaskMonitor;
import org.jdtaus.core.text.Message;

/**
 * jDTAUS Core SPI {@code TaskMonitor} reference implementation.
 * <p>The reference implementation uses a thread checking the state of all tasks
 * in the system periodically which is started upon initialization and runs
 * endlessly. Monitoring is controlled by property {@code pollIntervalMillis}
 * specifying the milliseconds of one period. Each time a period ends, tasks
 * are checked for state changes and corresponding events are fired. Property
 * {@code pollIntervalMills} defaults to {@code 250ms}.</p>
 *
 * <p><b>Note:</b><br/>
 * {@code TaskEvent}s of type {@code STARTED} and {@code ENDED} are fired by the
 * thread executing the task's operation. Since tasks are monitored
 * asynchronously, {@code TaskEvent}s of type {@code CHANGED_STATE} are fired by
 * the monitor thread, not by the thread executing the task's operation. Make
 * sure {@code TaskListener} implementations are prepared for being notified
 * by a different thread than the one executing a task's operation.</p>
 *
 * @author <a href="mailto:cs@schulte.it">Christian Schulte</a>
 * @version $Id: DefaultTaskMonitor.java 4365 2007-08-23 02:35:39Z schulte2005 $
 *
 * @see org.jdtaus.core.container.Container
 * @see #initialize()
 */
public final class DefaultTaskMonitor
    implements TaskMonitor, ContainerInitializer
{
    //--Implementation----------------------------------------------------------

// <editor-fold defaultstate="collapsed" desc=" Generated Code ">//GEN-BEGIN:jdtausImplementation
    // This section is managed by jdtaus-container-mojo.

    /** Meta-data describing the implementation. */
    private static final Implementation META =
        ModelFactory.getModel().getModules().
        getImplementation(DefaultTaskMonitor.class.getName());
// </editor-fold>//GEN-END:jdtausImplementation

    //----------------------------------------------------------Implementation--
    //--Constructors------------------------------------------------------------

// <editor-fold defaultstate="collapsed" desc=" Generated Code ">//GEN-BEGIN:jdtausConstructors
    // This section is managed by jdtaus-container-mojo.

    /**
     * <code>DefaultTaskMonitor</code> implementation constructor.
     *
     * @param meta Implementation meta-data.
     *
     * @throws NullPointerException if <code>meta</code> is <code>null</code>.
     */
    private DefaultTaskMonitor(final Implementation meta)
    {
        super();
        if(meta == null)
        {
            throw new NullPointerException("meta");
        }
        this.initializeProperties(meta.getProperties());
    }
    /**
     * <code>DefaultTaskMonitor</code> dependency constructor.
     *
     * @param meta dependency meta-data.
     *
     * @throws NullPointerException if <code>meta</code> is <code>null</code>.
     */
    private DefaultTaskMonitor(final Dependency meta)
    {
        super();
        if(meta == null)
        {
            throw new NullPointerException("meta");
        }
        this.initializeProperties(meta.getProperties());
    }

    /**
     * Initializes the properties of the instance.
     *
     * @param meta the property values to initialize the instance with.
     *
     * @throws NullPointerException if {@code meta} is {@code null}.
     */
    private void initializeProperties(final Properties meta)
    {
        Property p;

        if(meta == null)
        {
            throw new NullPointerException("meta");
        }

        p = meta.getProperty("pollIntervalMillis");
        this._pollIntervalMillis = ((java.lang.Long) p.getValue()).longValue();

    }
// </editor-fold>//GEN-END:jdtausConstructors

    //------------------------------------------------------------Constructors--
    //--ContainerInitializer----------------------------------------------------

    /**
     * Initializes the instance and starts a thread monitoring all tasks for
     * changes.
     *
     * @throws PropertyException for invalid property values.
     *
     * @see #assertValidProperties()
     */
    public void initialize()
    {
        this.assertValidProperties();
        this.monitorThread = new MonitorThread();
        this.monitorThread.start();
    }

    //----------------------------------------------------ContainerInitializer--
    //--Dependencies------------------------------------------------------------

// <editor-fold defaultstate="collapsed" desc=" Generated Code ">//GEN-BEGIN:jdtausDependencies
    // This section is managed by jdtaus-container-mojo.

    /** Configured <code>Logger</code> implementation. */
    private transient Logger _dependency0;

    /**
     * Gets the configured <code>Logger</code> implementation.
     *
     * @return the configured <code>Logger</code> implementation.
     */
    private Logger getLogger()
    {
        Logger ret = null;
        if(this._dependency0 != null)
        {
            ret = this._dependency0;
        }
        else
        {
            ret = (Logger) ContainerFactory.getContainer().
                getDependency(DefaultTaskMonitor.class,
                "Logger");

            if(ModelFactory.getModel().getModules().
                getImplementation(DefaultTaskMonitor.class.getName()).
                getDependencies().getDependency("Logger").
                isBound())
            {
                this._dependency0 = ret;
            }
        }

        if(ret instanceof ContextInitializer && !((ContextInitializer) ret).
            isInitialized(ContextFactory.getContext()))
        {
            ((ContextInitializer) ret).initialize(ContextFactory.getContext());
        }

        return ret;
    }
// </editor-fold>//GEN-END:jdtausDependencies

    //------------------------------------------------------------Dependencies--
    //--Properties--------------------------------------------------------------

// <editor-fold defaultstate="collapsed" desc=" Generated Code ">//GEN-BEGIN:jdtausProperties
    // This section is managed by jdtaus-container-mojo.

    /**
     * Property {@code pollIntervalMillis}.
     * @serial
     */
    private long _pollIntervalMillis;

    /**
     * Gets the value of property <code>pollIntervalMillis</code>.
     *
     * @return the value of property <code>pollIntervalMillis</code>.
     */
    public long getPollIntervalMillis()
    {
        return this._pollIntervalMillis;
    }

// </editor-fold>//GEN-END:jdtausProperties

    //--------------------------------------------------------------Properties--
    //--TaskEventSource---------------------------------------------------------

    /** List of {@code TaskListener}s. */
    private final EventListenerList taskListeners = new EventListenerList();

    public void addTaskListener(final TaskListener listener)
    {
        if(listener == null)
        {
            throw new NullPointerException("listener");
        }

        this.taskListeners.add(TaskListener.class, listener);
    }

    public void removeTaskListener(final TaskListener listener)
    {
        if(listener == null)
        {
            throw new NullPointerException("listener");
        }

        this.taskListeners.remove(TaskListener.class, listener);
    }

    public TaskListener[] getTaskListeners()
    {
        return (TaskListener[]) this.taskListeners.
            getListeners(TaskListener.class);

    }

    //---------------------------------------------------------TaskEventSource--
    //--TaskMonitor-------------------------------------------------------------

    public void monitor(final Task task)
    {
        if(task == null)
        {
            throw new NullPointerException("task");
        }

        synchronized(this.monitorThread)
        {
            if(!this.monitorThread.isAlive())
            { // Monitoring died for some reason; start a new thread.
                this.getLogger().warn(
                    DefaultTaskMonitorBundle.getThreadDiedText(
                    Locale.getDefault()));

                this.monitorThread = new MonitorThread();
                this.monitorThread.start();
            }
        }

        synchronized(this.stateMap)
        {
            this.createTaskState(task);
            this.fireTaskEvent(new TaskEvent(task, TaskEvent.STARTED));
        }
    }

    public void finish(final Task task)
    {
        if(task == null)
        {
            throw new NullPointerException("task");
        }

        synchronized(this.stateMap)
        {
            if(this.changedState(task))
            {
                this.fireTaskEvent(new TaskEvent(task,
                    TaskEvent.CHANGED_STATE));

            }

            this.removeTaskState(task);
            this.fireTaskEvent(new TaskEvent(task, TaskEvent.ENDED));
        }
    }

    //-------------------------------------------------------------TaskMonitor--
    //--DefaultTaskMonitor------------------------------------------------------

    /** Caches the state of a task. */
    private static final class TaskState
    {
        boolean indeterminate;
        boolean cancelable;
        boolean cancelled;
        int minimum;
        int maximum;
        int progress;
        Message progressDescription;
    }

    /** Thread monitoring all currently running {@code Task}s for changes. */
    private class MonitorThread extends Thread
    {
        /** Creates a new {@code MonitorThread} instance. */
        private MonitorThread()
        {
            super("DefaultTaskMonitor");
        }

        /** {@inheritDoc} */
        public void run()
        {
            while(true)
            {
                try
                {
                    Thread.sleep(getPollIntervalMillis());
                    this.checkTasks();
                }
                catch(InterruptedException e)
                {
                    this.checkTasks();
                }
            }
        }

        public void start()
        {
            super.start();
            getLogger().info(DefaultTaskMonitorBundle.getThreadStartedMessage(
                Locale.getDefault()).format(new Object[] {
                new Long(getPollIntervalMillis() )}));

        }

        /**
         * Checks the state of all currently running tasks for changes and
         * fires corresponding events.
         */
        private void checkTasks()
        {
            synchronized(stateMap)
            {
                for(Iterator it = stateMap.keySet().iterator(); it.hasNext();)
                {
                    final Task task = (Task) it.next();
                    if(changedState(task))
                    {
                        fireTaskEvent(new TaskEvent(task,
                            TaskEvent.CHANGED_STATE));

                    }
                }
            }
        }
    }

    /** Maps {@code Task}s to corresponding {@code TaskState} instances. */
    private final Map stateMap = new HashMap(1000);

    /** The thread monitoring tasks. */
    private MonitorThread monitorThread;

    /** Creates a new {@code DefaultTaskMonitor} instance. */
    public DefaultTaskMonitor()
    {
        this(META);
        this.initialize();
    }

    /**
     * Creates a new {@code DefaultTaskMonitor} instance taking the
     * milliseconds of one period.
     *
     * @param pollIntervalMillis milliseconds for one monitoring period.
     *
     * @throws PropertyException if {@code pollIntervalMillis} is less than
     * {@code 100ms}.
     */
    public DefaultTaskMonitor(final long pollIntervalMillis)
    {
        this(META);
        this._pollIntervalMillis = pollIntervalMillis;
        this.initialize();
    }

    /**
     * Checks configured properties.
     *
     * @throws PropertyException for illegal property values.
     */
    private void assertValidProperties()
    {
        if(this.getPollIntervalMillis() < 100L)
        {
            throw new PropertyException("pollIntervalMillies",
                Long.toString(this.getPollIntervalMillis()));

        }
    }

    /**
     * Notifies all registered {@code TaskListener}s about {@code TaskEvent}s.
     *
     * @param e The event to be provided to the listeners.
     */
    private void fireTaskEvent(final TaskEvent e)
    {
        if(e == null)
        {
            throw new NullPointerException("e");
        }


        final Object[] listeners = this.taskListeners.getListenerList();
        for (int i = listeners.length - 2; i >= 0; i -= 2)
        {
            if (listeners[i] == TaskListener.class)
            {
                ((TaskListener) listeners[i + 1]).onTaskEvent(e);
            }
        }
    }

    /**
     * Caches the state of a {@code Task}.
     *
     * @param task the task to cache state for.
     *
     * @throws NullPointerException if {@code task} is {@code null}.
     * @throws IllegalStateException if the cache already holds state for
     * {@code task}.
     */
    private void createTaskState(final Task task)
    {
        final TaskState state = new TaskState();
        state.cancelable = task.isCancelable();
        state.indeterminate = task.isIndeterminate();
        state.cancelled = state.cancelable ? task.isCancelled() : false;
        state.progressDescription = task.getProgressDescription();

        if(state.indeterminate)
        {
            state.maximum = Integer.MIN_VALUE;
            state.minimum = Integer.MIN_VALUE;
            state.progress = Integer.MIN_VALUE;
        }
        else
        {
            state.maximum = task.getMaximum();
            state.minimum = task.getMinimum();
            state.progress = task.getProgress();
        }

        if(this.stateMap.put(task, state) != null)
        {
            throw new IllegalStateException(DefaultTaskMonitorBundle.
                getTaskAlreadyStartedMessage(Locale.getDefault()).
                format(new Object[] {
                task.getDescription().getText(Locale.getDefault()),
                new Date(task.getTimestamp()) }));

        }
    }

    /**
     * Removes the cached state of a {@code Task}.
     *
     * @param task the task to remove the cached state of.
     *
     * @throws NullPointerException if {@code task} is {@code null}.
     */
    private void removeTaskState(final Task task)
    {
        if(task == null)
        {
            throw new NullPointerException("task");
        }

        this.stateMap.remove(task);
    }

    /**
     * Checks the state of a given task for changes.
     *
     * @param task the task to check for state changes.
     *
     * @return {@code true} if the state of {@code task} changed since the last
     * time this method got called; {@code false} if the state did not change.
     *
     * @throws NullPointerException if {@code task} is {@code null}.
     * @throws IllegalStateException if no cached state exists for {@code task}.
     */
    private boolean changedState(final Task task)
    {
        if(task == null)
        {
            throw new NullPointerException("task");
        }


        boolean changedState = false;
        final TaskState state = (TaskState) this.stateMap.get(task);

        if(state == null)
        {
            throw new IllegalStateException();
        }

        if(state.indeterminate)
        {
            state.indeterminate = task.isIndeterminate();
            if(!state.indeterminate)
            {
                state.minimum = task.getMinimum();
                state.maximum = task.getMaximum();
                state.progress = task.getProgress();
                changedState = true;
            }
        }
        else
        {
            state.indeterminate = task.isIndeterminate();
            if(state.indeterminate)
            {
                changedState = true;
            }
            else
            {
                if(state.minimum != task.getMinimum())
                {
                    state.minimum = task.getMinimum();
                    changedState = true;
                }
                if(state.maximum != task.getMaximum())
                {
                    state.maximum = task.getMaximum();
                    changedState = true;
                }
                if(state.progress != task.getProgress())
                {
                    state.progress = task.getProgress();
                    changedState = true;
                }
            }
        }

        if(state.cancelable)
        {
            state.cancelable = task.isCancelable();
            if(!state.cancelable)
            {
                changedState = true;
            }
            else
            {
                if(state.cancelled != task.isCancelled())
                {
                    state.cancelled = task.isCancelled();
                    changedState = true;
                }
            }
        }
        else
        {
            state.cancelable = task.isCancelable();
            if(!state.cancelable)
            {
                state.cancelled = false;
                changedState = true;
            }
        }

        if(state.progressDescription != task.getProgressDescription())
        {
            state.progressDescription = task.getProgressDescription();
            changedState = true;
        }
        else if(state.progressDescription != null &&
            !state.progressDescription.getText(Locale.getDefault()).
            equals(task.getProgressDescription().getText(
            Locale.getDefault())))
        {
            changedState = true;
        }

        return changedState;
    }

    //------------------------------------------------------DefaultTaskMonitor--
}
